module balance
  use precision
  use ll_intervals
  use eigen_toep
  use toeplitz_cauchy
  
  implicit none

  !include 'mpif.h'

  public :: seq_balance, seq_balance_levdur !, mpi_balance, mpi_balance_scala
  private

contains
  
  integer function seq_balance(ll_interv, n, T, lbg, ubg, max_eig, eigs_prev_, eigs_in_, eigs_post_)

    type(ll_list), intent(inout) :: ll_interv
    integer, intent(in) :: n, max_eig
    integer, intent(in), optional :: eigs_prev_, eigs_in_, eigs_post_
    real(kind=r_kind), intent(in) :: lbg, ubg, T(n)
  
    integer :: eigs_prev, eigs_in, eigs_post
    integer :: j, ret, n_eigs, afegits, pos, neg, n_intervs
    real(kind=r_kind) :: lb, ub, shift, eps, C(n)

    eps = epsilon(lb)

    if (.not. present(eigs_prev_)) then
       C = T
       C(1) = C(1) - lbg

       eigs_prev = dtsysv_neg(C)
    else
       eigs_prev = eigs_prev_
    end if

    if (.not. present(eigs_post_)) then 
       C = T
       C(1) = C(1) - ubg

       eigs_post = n - dtsysv_neg(C)
    else
       eigs_post = eigs_post_
    end if

    if (.not. present(eigs_in_)) then
       eigs_in = n - (eigs_prev+eigs_post)
    else
       eigs_in = eigs_in_
    end if

    !print *, 'abans init', eigs_in
    call ll_init(ll_interv, lbg, ubg, eigs_in, eigs_prev, eigs_post)
    ret = ll_i_node(ll_interv, 1, lb, ub, n_eigs)
    !print *,'primer estat', lb, ub, n_eigs

    
    j=1
    n_intervs = ll_length(ll_interv)
    do while (j<=n_intervs)       
       ret = ll_i_node(ll_interv, j, lb, ub, n_eigs)
       !print *,'processant estat',j, lb, ub, n_eigs, ub-lb
       if (ret .lt. 0) then
          seq_balance = -1
          return
       end if

  !!     print *, 'estat', lb, ub, ub-lb, eps
       
       if (n_eigs .le. max_eig) then
          j = j + 1
          afegits = 0
       elseif (ub-lb .lt. 100000*eps) then
          ! Hem trobat un clúster
          j = j + 1
          afegits = 0
       else
          shift = (lb+ub)/2
          C = T
          C(1) = C(1) - shift
          
          neg = dtsysv_neg(C)
          pos = n - neg
          
          afegits = ll_split_node(ll_interv, j, neg, pos)
          if (afegits .lt. 0) then
             seq_balance = -2
             return
          end if

!          print *, afegits, shift, neg, pos
          
       end if
       
!       print *, 'afegits:', afegits
       !print *

       n_intervs = ll_length(ll_interv)
    end do
  
    seq_balance = 0
    
  end function seq_balance
  
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  integer function seq_balance_levdur(ll_interv, n, T, lbg, ubg, max_eig, eigs_prev_, eigs_in_, eigs_post_)

    type(ll_list), intent(inout) :: ll_interv
    integer, intent(in) :: n, max_eig
    integer, intent(in), optional :: eigs_prev_, eigs_in_, eigs_post_
    real(kind=r_kind), intent(in) :: lbg, ubg, T(:)
  
    integer :: eigs_prev, eigs_in, eigs_post
    integer :: j, ret, n_eigs, afegits, pos, neg, n_intervs
    real(kind=r_kind) :: lb, ub, shift, eps, C(n)

    eps = epsilon(lb)

    if (.not. present(eigs_prev_)) then
       C = T
       C(1) = C(1) - lbg

       eigs_prev = dtsy_neg_lev(n,C)
    else
       eigs_prev = eigs_prev_
    end if

    if (.not. present(eigs_post_)) then 
       C = T
       C(1) = C(1) - ubg

       eigs_post = n - dtsy_neg_lev(n,C)
    else
       eigs_post = eigs_post_
    end if

    if (.not. present(eigs_in_)) then
       eigs_in = n - (eigs_prev+eigs_post)
    else
       eigs_in = eigs_in_
    end if

!    print *, 'abans init', eigs_in
    call ll_init(ll_interv, lbg, ubg, eigs_in, eigs_prev, eigs_post)
    ret = ll_i_node(ll_interv, 1, lb, ub, n_eigs)
!    print *,'primer estat', lb, ub, n_eigs

    
    j=1
    n_intervs = ll_length(ll_interv)
    do while (j<=n_intervs)       
       ret = ll_i_node(ll_interv, j, lb, ub, n_eigs)
       !print *,'processant estat',j, lb, ub, n_eigs, ub-lb
       if (ret .lt. 0) then
          seq_balance_levdur = -1
          return
       end if

  !!     print *, 'estat', lb, ub, ub-lb, eps
       
       if (n_eigs .le. max_eig) then
          j = j + 1
          afegits = 0
       elseif (ub-lb .lt. 100000*eps) then
          ! Hem trobat un clúster
          j = j + 1
          afegits = 0
       else
          shift = (lb+ub)/2
          C = T
          C(1) = C(1) - shift
          
          neg = dtsy_neg_lev(n,C)
          pos = n - neg
          
          afegits = ll_split_node(ll_interv, j, neg, pos)
          if (afegits .lt. 0) then
             seq_balance_levdur = -2
             return
          end if

!          print *, afegits, shift, neg, pos
          
       end if
       
!       print *, 'afegits:', afegits
       !print *

       n_intervs = ll_length(ll_interv)
    end do
  
    seq_balance_levdur = 0
    
  end function seq_balance_levdur
  
end module balance
