!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
! Para que funcione el program principal tiene que leer 
! la variable de entorno del modulo threadsnumber y 
! darle valor a la variable num_threads
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
module cauchy_inertia
  use precision
  use blocksize
  use threadsnumber
  use iso_c_binding
  use pthreads
  implicit none

  public :: cauchy_neg, inertia_time

     real( kind=r_kind ) :: tiempo_inertia = 0

  private 
    integer( c_int ), target, dimension(128) :: cond_queue_there_are_tasks
    integer( c_int ), target, dimension(128) :: task_queue_cond_lock
    integer( c_int ), target, dimension(128) :: attr
    !integer( c_int ), target, allocatable, dimension(:) :: cond_queue_there_are_tasks
    !integer( c_int ), target, allocatable, dimension(:) :: task_queue_cond_lock
    !integer( c_int ), target, allocatable, dimension(:) :: attr

    integer :: n, blocks, neg

    logical, dimension(:,:), allocatable :: task_available

    real(kind=r_kind), dimension(:), pointer :: g1_pointer, g2_pointer, diagC_pointer, lambda_pointer
    
    real( kind=r_kind ) :: t1, t2, tn

    ! This module uses variables 
    !   1. block_size defined in module blocksize
    !   2. num_threads defined in module threadsnumber

  contains
  
  integer function cauchy_neg( lambda, diagC, g1, g2 ) 
    real(kind=r_kind), dimension(:)  , intent(in)   , target :: lambda
    real(kind=r_kind), dimension(:)  , intent(inout), target :: diagC, g1, g2

    ! Computes the triangular LDLT decomposition of a symmetric Cauchy matrix C=L*D*L'
    ! Pay attention the signature matrix H = [ [ 0 1 ];[ -1 0 ] ]

    integer :: id
    integer :: status

    integer( c_int ) :: rc
    type( pthread_t ), target, allocatable :: thread_id(:)
    integer( c_int ), target, allocatable :: tid(:) 

    type( c_ptr ), target :: c_pointer = C_NULL_PTR

    neg = 0

    !print *,'cauchy_neg . num_threads = ',num_threads
    if( num_threads.le.0 ) then 
      print *,'Num threads not valid = ',num_threads
      stop
    end if
    n = size(lambda,1)
    if( n.ne.size(diagC,1) ) return 
    if( n.ne.size(g1,1) )    return 
    if( n.ne.size(g2,1) )    return 
    allocate( thread_id(num_threads), tid(num_threads), Stat=status )
    if( status.ne.0 ) then 
      print *,'Error in tids array allocation( status = ',status,')'
      stop
    end if

    ! Assigning pointers
    g1_pointer => g1
    g2_pointer => g2
    diagC_pointer => diagC
    lambda_pointer => lambda

    blocks = n/block_size
    if( mod(n,block_size).gt.0 ) blocks = blocks + 1
    ! Initializing tasks matrix to FALSE
    allocate( task_available(blocks,blocks), Stat=status )
    if( status.ne.0 ) then 
      print *,'Error in tasks allocation( status = ',status,')'
      stop
    end if
    task_available = .false.
    !print *,'n = ', n
    !print *,'blocks = ', blocks
    !print *,task_available

    !allocate( task_queue_cond_lock(64), cond_queue_there_are_tasks(64), attr(64), Stat=status )
    !print *,'1. Por aqui'
    rc = pthread_attr_init( c_loc( attr ) )
    if( rc.ne.0 ) then 
      print *,'Error in pthread_attr_init ( rc = ',rc,')'
      stop
    end if
    !print *,'2. Por aqui'
    rc = pthread_attr_setdetachstate( c_loc( attr ), PTHREAD_CREATE_JOINABLE )
    if( rc.ne.0 ) then 
      print *,'Error in pthread_attr_setdetachstate ( rc = ',rc,')'
      stop
    end if
    !print *,'3. Por aqui'
    rc = pthread_mutex_init( c_loc( task_queue_cond_lock ), C_NULL_PTR ) 
    if( rc.ne.0 ) then 
      print *,'Error in pthread_mutex_init ( rc = ',rc,')'
      stop
    end if
    !print *,'4. Por aqui'
    rc = pthread_cond_init( c_loc( cond_queue_there_are_tasks ),  C_NULL_PTR )
    if( rc.ne.0 ) then 
      print *,'Error in pthread_cond_init ( rc = ',rc,')'
      stop
    end if
    !print *,'n = ', n
    !print *,'blocks = ', blocks
    !print *,task_available

    !CAMBIAR EL PUNTERO A LA FNCION POR c_funloc(thread_routine)
    ! Creating threads
    !print *,'Creating threads'
    do id = 1, num_threads
      tid(id) = id-1
      rc = pthread_create( c_loc(thread_id(id)), C_NULL_PTR, c_funloc(thread_routine), c_loc(tid(id)) ) 
      if( rc.ne.0 ) then 
        print *,'Error in creating threads ( rc = ',rc,')'
      !  stop
      end if
      !print *,'Created thread ',id,' (',thread_id(id),')'
    end do

    !print *,'Joining threads '
 
    do id = 1, num_threads
      !print *,'joining thread ',id, ' (',thread_id(id),')'
      !rc = pthread_join( thread_id(id), c_loc( C_NULL_PTR ) )
      rc = pthread_join( thread_id(id), c_loc( c_pointer ) )
      if( rc.ne.0 ) then 
        print *,'Error in joining threads ( rc = ',rc,')'
      !  stop
      end if
      !print *,'thread ',id,' joined'
    end do
    !print *,'threads joined'

    rc = pthread_mutex_destroy( c_loc( task_queue_cond_lock ) )
    if( rc.ne.0 ) then 
      print *,'Error in mutex destroy ( rc = ',rc,')'
      stop
    end if
    rc = pthread_cond_destroy( c_loc( cond_queue_there_are_tasks ) )
    if( rc.ne.0 ) then 
      print *,'Error in cond destroy ( rc = ',rc,')'
      stop
    end if

    deallocate( thread_id, Stat=status )
    if( status.ne.0 ) then 
      print *,'Error deallocating thread_id ( status = ',status,')'
      stop
    end if
    deallocate( tid, Stat=status )
    if( status.ne.0 ) then 
      print *,'Error deallocating tid ( status = ',status,')'
      stop
    end if
    deallocate( task_available, Stat=status )
    if( status.ne.0 ) then 
      print *,'Error deallocating task_available ( status = ',status,')'
      stop
    end if
    !deallocate( task_queue_cond_lock, cond_queue_there_are_tasks, attr, Stat=status )

    cauchy_neg = neg

  end function cauchy_neg
    
  subroutine thread_routine( c_id_pointer ) 
    type( c_ptr ), value :: c_id_pointer

    integer, pointer :: id_pointer
    integer :: tid
    integer :: i, j
    integer( c_int ) :: rc

    call c_f_pointer( c_id_pointer, id_pointer )
    tid = id_pointer

    do i = 1, blocks
      if( tid.eq.mod(i-1,num_threads) ) then
        do j = 1, i-1
          rc = pthread_mutex_lock( c_loc( task_queue_cond_lock ) )
          do while ( .not.task_available(i,j) )
            rc = pthread_cond_wait( c_loc( cond_queue_there_are_tasks ), c_loc( task_queue_cond_lock ) )
          end do
          rc = pthread_mutex_unlock( c_loc( task_queue_cond_lock ) )
          call process_task( i, j )
        end do
        call process_task( i, i )
        rc = pthread_mutex_lock( c_loc(task_queue_cond_lock) )
        do j=i+1, blocks
          task_available(j,i) = .true.
        end do
        rc = pthread_cond_broadcast( c_loc( cond_queue_there_are_tasks ) )
        rc = pthread_mutex_unlock( c_loc( task_queue_cond_lock ) )
      end if
    end do

  end subroutine thread_routine

  subroutine process_task( i, j )
    integer, intent(in) :: i, j
    integer :: initblk, endblk, bsize
    integer :: initblk1, endblk1

    !print *,'Processing task (',i,',',j,')'
    initblk = (i-1)*block_size+1
    bsize   = block_size
    if( i == blocks .and. mod(n,block_size).ne.0 ) bsize = mod(n,block_size)
    endblk  = initblk+bsize-1

    if( i.eq.j ) then
      !print *,'Processing task (',i,',',j,')',' --> triang = g1_pointer(',initblk,':',endblk,')='!,g1_pointer(initblk:endblk)
      neg = neg + triang( g1_pointer(initblk:endblk), g2_pointer(initblk:endblk),            & 
                          diagC_pointer(initblk:endblk), lambda_pointer(initblk:endblk) )
    else 
      !print *,'Processing task (',i,',',j,')',' --> square = g1_pointer(',initblk,':',endblk,')='!,g1_pointer(initblk:endblk)
      initblk1 = (j-1)*block_size+1
      endblk1  = initblk1+block_size-1
      !print *,'Processing task (',i,',',j,')',' --> square = g1_pointer(',initblk1,':',endblk1,')='!,g1_pointer(initblk:endblk)
      call square( g1_pointer(initblk:endblk), g2_pointer(initblk:endblk),            & 
                   diagC_pointer(initblk:endblk), lambda_pointer(initblk:endblk),     &
                   g1_pointer(initblk1:endblk1), g2_pointer(initblk1:endblk1),        & 
                   diagC_pointer(initblk1:endblk1), lambda_pointer(initblk1:endblk1) )
    end if
  end subroutine process_task

  integer function triang( g1, g2, diagC, lambda ) 
    real(kind=r_kind), dimension(:), intent(in) :: lambda
    real(kind=r_kind), dimension(:), intent(inout) :: diagC, g1, g2

    integer :: n
    integer :: i, k
    integer :: neg_values 
    real(kind=r_kind), dimension(BLK_ITANIUM_SIZE) :: Glocal1, Glocal2

    real(kind=r_kind) :: d, gk1, gk2, gi1, gi2, lambdak, aux, lik, eps

    n = size( lambda, 1 )
    neg_values = 0
    if( n.ne.size(diagC,1) .or.  n.ne.size(g1,1) .or.  n.ne.size(g2,1) ) then
      print *,'Error in dimensions in triang'
      stop
    end if

    if( n.gt.BLK_ITANIUM_SIZE ) then
      print *,'Error in triang (nb>BLK_ITANIUM_SIZE)'
      stop
    end if

    eps = epsilon(d)

    do k = 1, n
      Glocal1(k) = g1(k)
      Glocal2(k) = g2(k)
    end do

    do k=1, n-1
       d = diagC(k)
       if( d.le.eps*1000 ) neg_values = neg_values + 1
       gk1 = Glocal1(k)
       gk2 = Glocal2(k)
       lambdak = lambda(k)

       do i=k+1, n
          gi1 = Glocal1(i)
          gi2 = Glocal2(i)
          aux = (-gi2*gk1 + gi1*gk2) / (lambda(i) - lambdak)
          lik = aux/d
          gi1 = gi1 - gk1 * lik
          gi2 = gi2 - gk2 * lik
          Glocal1(i) = gi1
          Glocal2(i) = gi2
          diagC(i) = diagC(i) - ( aux*lik )
       end do 

    end do
    if( diagC(n).le.eps*1000 ) neg_values = neg_values + 1

    do k=1, n
      g1(k) = Glocal1(k)
      g2(k) = Glocal2(k)
    end do

    triang = neg_values

  end function triang

  subroutine square ( g1, g2, diagC, lambda, gg1, gg2, diagC1, lambda1 )
    real(kind=r_kind), dimension(:), intent(in) :: lambda, lambda1
    real(kind=r_kind), dimension(:), intent(inout) :: g1, g2, gg1, gg2, diagC, diagC1

    integer :: n, nb
    integer :: h, i, j, k
    integer :: sizec
    real(kind=r_kind) :: d, gk1, gk2, gi1, gi2, lambdak, aux, lik
    real(kind=r_kind), parameter :: BLK_COL_SIZE=4
    real(kind=r_kind), dimension(BLK_ITANIUM_SIZE) :: Glocal1, Glocal2

    !call ctimer( t1, tn, tn )
    n = size( lambda, 1 )
    if( n.ne.size(diagC,1) .or. n.ne.size(g1,1) .or. n.ne.size(g2,1) ) then
      print *,'Error in dimensions in square'
      stop
    end if

    nb = size( lambda1, 1 )
    if( nb.ne.size(diagC1,1) .or. nb.ne.size(gg1,1) .or. nb.ne.size(gg2,1) ) then
      print *,'Error in dimensions in square'
      stop
    end if

    if( n.gt.BLK_ITANIUM_SIZE ) then
      print *,'Error in square (nb>BLK_ITANIUM_SIZE)'
      stop
    end if

    Glocal1 = g1
    Glocal2 = g2

    do j=1, nb, BLK_COL_SIZE !runs over columns
      sizec = nb-j+1;
      if( sizec>BLK_COL_SIZE ) sizec = BLK_COL_SIZE
      k = j
      do h=1, sizec
        d = diagC1(k)
        lambdak = lambda1(k)
        gk1 = gg1(k)
        gk2 = gg2(k)
  
        do i = 1, n
          gi1 = Glocal1(i)
          gi2 = Glocal2(i)
          aux = (-gi2*gk1 + gi1*gk2) / (lambda(i) - lambdak)
          lik = aux/d
          gi1 = gi1 - gk1 * lik
          gi2 = gi2 - gk2 * lik
          Glocal1(i) = gi1
          Glocal2(i) = gi2
          diagC(i) = diagC(i) - ( (aux*aux)/d )
        end do
        k = k + 1
      end do 
    end do 

    g1 = Glocal1
    g2 = Glocal2

    !call ctimer( t2, tn, tn )
    tiempo_inertia = tiempo_inertia + (t2-t1)
  end subroutine square

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  function inertia_time
    real(kind=r_kind) :: inertia_time

    inertia_time=tiempo_inertia
  end function inertia_time
  
end module cauchy_inertia

