module cl_args
  use precision
  implicit none

  public :: cl_get
  private :: cl_get_int, cl_get_real, cl_get_char

  interface cl_get
     module procedure cl_get_int, cl_get_real, cl_get_char
  end interface

contains

  integer function cl_get_int(tag, value)
    character(*), intent(in) :: tag
    integer, intent(out)     :: value

    integer       :: n_args, i, status
    character(64) :: buffer

    n_args = iargc()

    do i=1,n_args
       call getarg(i, buffer, status)
       if (status .lt. 0) then
          cl_get_int = -1
          return
       end if
       
       if (buffer(:status) .eq. tag) then
          call getarg(i+1, buffer, status)
          if (status .lt. 0) then
             cl_get_int = -2
             return
          end if
          
          read (buffer, '(i)', iostat=status) value
          if (status .ne. 0) then
             cl_get_int = -3
             return
          end if

          cl_get_int = 0
          return
       end if
       
    end do
    
    cl_get_int = -1
    return
    
  end function cl_get_int

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  integer function cl_get_real(tag, value)
    character(*), intent(in) :: tag
    real(kind=r_kind), intent(out)     :: value

    integer       :: n_args, i, status
    character(64) :: buffer

    n_args = iargc()

    do i=1,n_args
       call getarg(i, buffer, status)
       if (status .lt. 0) then
          cl_get_real = -1
          return
       end if
       
       if (buffer(:status) .eq. tag) then
          call getarg(i+1, buffer, status)
          if (status .lt. 0) then
             cl_get_real = -2
             return
          end if
          
          read (buffer, '(f20.0)', iostat=status) value
          if (status .ne. 0) then
             cl_get_real = -3
             return
          end if
             
          cl_get_real = 0
          return
       end if
       
    end do
    
    cl_get_real = -1
    return
    
  end function cl_get_real


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  integer function cl_get_char(tag, value)
    character(*), intent(in) :: tag
    character*1, intent(out) :: value

    integer       :: n_args, i, status
    character(64) :: buffer

    n_args = iargc()

    do i=1,n_args
       call getarg(i, buffer, status)
       if (status .lt. 0) then
          cl_get_char = -1
          return
       end if
       
       if (buffer(:status) .eq. tag) then
          call getarg(i+1, buffer, status)
          if (status .lt. 0) then
             cl_get_char = -2
             return
          end if
          
          read (buffer, '(a)', iostat=status) value
          if (status .ne. 0) then
             cl_get_char = -3
             return
          end if
             
          cl_get_char = 0
          return
       end if
       
    end do
    
    cl_get_char = -1
    return
    
  end function cl_get_char
  
end module cl_args
