module conv_lib
  use precision   ! exports kind value r_kind
  use rft_mkl     
  
  implicit none
  
  public :: conv
  private

contains
  
  subroutine conv( t, x, y, u_ ) 
    
    !     conv performs the product T*x by a convolution operation
    !
    real(kind=r_kind), dimension(:), intent(in)           :: t, x
    real(kind=r_kind), dimension(:), intent(in), optional :: u_
    real(kind=r_kind), dimension(:), intent(out)          :: y

    real(kind=r_kind), dimension(:), allocatable :: rft_t, rft_x

    integer :: m, n
    logical :: non_symmetric
    
    integer :: o, info, i
    !real(kind=r_kind) :: r

    m = size(t,1)
    non_symmetric = present( u_ )
    
    if( non_symmetric ) then
      n = size(u_,1)
      if( size(x,1).ne.n .or. size(y,1).ne.m ) then
        print *,'conv: Dimensions error'
        return
      end if
    end if

    if( non_symmetric ) then
      print *,'conv: Non symmetric convolution still not implemented'
      return
    else 
      o = 2
      do while( o.lt.(2*m-1) ) 
        o = 2 * o
      end do
      allocate( rft_t(o), rft_x( o ), stat=info)
      if (info .ne. 0) then
         print *,'Error allocating memory in conv'
         return
      end if      
      rft_t = (/t,(ZERO,i=0,o-2*m),t(m:2:-1)/)
      call rft( rft_t )
      !print *,'rft_t = ',rft_t
      rft_x = (/x,(ZERO,i=1,o-m)/)
      call rft( rft_x )
      !print *,'rft_x = ',rft_x
      !product of arrays
      rft_t(3:o:2) = rft_t(2:o:2)
      rft_t = rft_t * rft_x
      !print *,'rft_t = ',rft_t
      !rft_t(1) = rft_t(1) * rft_x(1)
      !do i = 2, o-1, 2
      !  r = rft_t(i)
      !  !print *,'rft_t(',i,')=',r
      !  !print *,'rft_x(',i,')=',rft_x(i)
      !  !print *,'rft_x(',i+1,')=',rft_x(i+1)
      !  rft_t( i )   = r * rft_x( i )
      !  rft_t( i+1 ) = r * rft_x( i+1 )
      !end do
      !rft_t(o) = rft_t(o) * rft_x(o)
      !print *,'rft_t = ',rft_t
      call irft( rft_t )
      y = rft_t(1:m) / o

      deallocate( rft_t, rft_x, stat=info)
      
    end if

    return
    
  end subroutine conv
  
end module conv_lib
