module eigen_toep
  use precision
  use m_mrgrnk
  use dst_lib
  use toeplitz_cauchy
  use diagCauchy
  !use cauchy_deco_solve
  implicit none

  !public :: dtsyeig, dtsyeig_arp, dtsyeig_newton, dtsy_norminf, dtsy_eigen_bounds, dtsy_neg, dtsy_sp2dns, dtsy_neg_lev, dtsyeig_nlanc
  public :: dtsyeig, dtsy_norminf, dtsy_eigen_bounds, dtsy_sp2dns, dtsy_neg_lev, dtsyeig_nlanc, eigen_toep_time, tiempo_eigen_toep

  real( kind = r_kind ) :: tiempo_eigen_toep = ZERO

  private

contains
  
  function dtsy_norminf(t)
    !      dtsty_norminf calcula la norma infinita d'una 
    ! matriu Toeplitz simetrica en format comprimit. En 
    ! aquest cas particular la norma infinita coincidix 
    ! amb norma 1.
    real(kind=r_kind) :: dtsy_norminf
    real(kind=r_kind), intent(in) :: t(:)

    integer :: n, mitat, i
    real(kind=r_kind) :: norm, max

    n = size(t,1)

    if (n .eq. 1) then
       dtsy_norminf = t(1)
       return
    end if
    
    mitat = n/2 + mod(n,2)
    norm = sum(abs(t))
    max = norm

    do i = 1,mitat
       norm = norm - abs(t(n-i+1)) + abs(t(i+1))
       if (norm .gt. max) max = norm
    end do
    
    dtsy_norminf = max
    
  end function dtsy_norminf
  
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  
  subroutine dtsy_sp2dns(n, vT, T)
   integer, intent(in) :: n
    real(kind=r_kind), intent(in)  :: vT(:) 
    real(kind=r_kind), intent(out) :: T(:,:)
    
    integer :: i, j, k
    
    do i=1,n
       k=1
       do j=i,2,-1
          T(k,i) = vT(j)
          k = k + 1
       end do
      
       do j=1,n-i+1
          T(k,i) = vT(j)
          k = k + 1
       end do
       
    end do
   
  end subroutine dtsy_sp2dns
  
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


  subroutine dtsy_eigen_bounds(t,lb,ub)
    real(kind=r_kind), intent(in) :: t(:)
    real(kind=r_kind), intent(out) :: lb, ub

    real(kind=r_kind) :: max_radio

    max_radio = dtsy_norminf(t) - t(1)

    lb = t(1) - max_radio
    ub = t(1) + max_radio

  end subroutine dtsy_eigen_bounds

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    integer function cauchy_neg_old(n,G,lambda,diagC)
    ! Input/Output parameters
    integer, intent(in) :: n
    real(kind=r_kind), intent(in) :: lambda(:)
    real(kind=r_kind), intent(inout) :: diagC(:)
    real(kind=r_kind), intent (inout) :: G(:,:)

    ! Local variables
    integer :: i, j, k, ierr, neg
    real(kind=r_kind) :: aux, d, lambdak, lik, gk1, gk2, eps
 
    ! Executable statements
    eps = epsilon(d)
    neg = 0

    do k=1,n
       lambdak = lambda(k)
       d = diagC(k)
       if (d .le. eps*1000) neg = neg + 1
       gk1 = G(k,1)
       gk2 = G(k,2)
       do i=k+1,n
          aux = (-G(i,2)*gk1 + G(i,1)*gk2) / (lambda(i) - lambdak)
          lik = aux/d
          G(i,1) = G(i,1) - gk1 * lik
          G(i,2) = G(i,2) - gk2 * lik
          diagC(i) = diagC(i) - lik*aux ! diagC(i) -( (aux**2)/d )
       end do
    end do

    cauchy_neg_old = neg
    
  end function cauchy_neg_old
  
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  subroutine Pdegrmsc(r,v,mv,nv,y,h)
    real(kind=r_kind), external :: dnrm2
    external dgemv, dcopy

    integer, intent(in)            :: mv, nv
!    real(kind=r_kind), intent(in)  :: r(mv), v(mv,nv)
!    real(kind=r_kind), intent(out) :: y(mv), h(nv+1)
    real(kind=r_kind), intent(in)  :: r(:), v(:,:)
    real(kind=r_kind), intent(out) :: y(:), h(:)


    integer, parameter :: maxort = 4
    real(kind=r_kind), parameter :: qmax = 0.7_r_kind
    integer :: nr, iort, ierr
!    real(kind=r_kind) :: hc(nv+1), hv(nv+1), raux(mv)
    real(kind=r_kind), allocatable :: hc(:), hv(:), raux(:)
    real(kind=r_kind) :: rnn, rnorm

    allocate(hc(nv+1), hv(nv+1), raux(mv), stat=ierr)
    if (ierr .ne. 0) stop "Error assignant memoria a l'entrar"
    
    rnn = dnrm2(mv,r,1)

    !raux = r
    call dcopy(mv, r, 1, raux, 1) 
    hc = 0_r_kind

    if (nv .gt. 0) then
       do iort=1,maxort
          !hv=matmul(transpose(v),raux)
          call dgemv('t',mv,nv,dble(1),v,mv,raux,1,dble(0),hv,1)

          hc = hc + hv

          !raux = raux - matmul(v(:,1:nv) , hv(1:nv))          
          call dgemv('n',mv,nv,dble(-1),v,mv,hv,1,dble(1),raux,1) 

          rnorm = rnn
          rnn = dnrm2(mv,raux,1)
          if (rnn .gt. (qmax * rnorm)) exit
       end do
    end if

    h(1:nv ) = hc
    h(nv+1) = rnn
    y = raux * (1/rnn) 

    deallocate(hc, hv, raux)

  end subroutine pdegrmsc


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


  subroutine dtsyeig (n, Toe, lb, ub, xv, lmb, iresult, tolconv_, jmax_, maxmul_)
    !DTSYEIG Solve eigenvalue problem with MATLAB's SPTARN template.
    ! Matrix A is symmetric Toeplitz
    ! Matriz B is identity
    !
    !       [XV,LMB,IRESULT]=SPTARN(A,B,LB,UB,SPD,TOLCONV,JMAX,MAXMUL)
    !       Finds eigenvalues of pencil (A-LMB*B)*XV=0 in interval LB < LMB <= UB.
    !
    !       Input data:
    !       A, B sparse matrices
    !       LB lower bound if -Inf all eigenvalues to the left of UB sought
    !       UB upper bound if +Inf all eigenvalues to the right of LB sought
    !       One of LB, UB must be finite. Narrower interval gives faster result.
    !       In complex case real parts of LMB are compared to LB and UB.
    !
    !       Output results:
    !       XV eigenvectors, ordered so that NORM(A*XV-B*XV*DIAG(LMB)) small
    !       LMB eigenvalues in interval, sorted
    !       IRESULT flag ABS(IRESULT) number of eigenvalues found
    !       IRESULT =>0 successful termination, all eigenvalues in interval found
    !       IRESULT < 0 not yet successful, there may be more eigenvalues. Try
    !       again with a smaller interval!
    !
    !       Parameters that may be set, but have defaults:
    !       SPD   true if pencil known to be symmetric pos definite
    !       default false
    !
    !       TOLCONV   expected relative accuracy
    !       default 100*eps machine precision
    !
    !       JMAX  maximum number of basis vectors. You need JMAX*N working space
    !       so a small value may be justified on a small computer, otherwise
    !       let it be the default value JMAX=100. Normally the algorithm stops
    !       earlier when enough eigenvalues have converged.
    !
    !       MAXMUL number of Arnoldi runs tried. Must at least be as large
    !       as maximum multiplicity of any eigenvalue. If a small value of
    !       JMAX is given, many Arnoldi runs are necessary.
    !       Default value is MAXMUL=N, which will be needed when all the
    !       eigenvalues of the unit matrix are sought!
    !
    !       Algorithm description:
    !       Uses Arnoldi with spectral transformation. The shift is chosen
    !       at UB, LB or at a random point in the interval (LB,UB) when both
    !       bounds are finite.
    !       Number of steps J in the Arnoldi run depends on how many
    !       eigenvalues there are in the interval, but stops at J=MIN(JMAX,N).
    !       After a
    !       stop, restart to find more Schur vectors in orthogonal complement to
    !       all those already found. When no more eigenvalues are found in
    !       LB < LMB <= UB, algorithm stops. For small values of JMAX, several
    !       restarts may be needed before a certain eigenvalue has converged,
    !       algorithm works for JMAX > #eigenvalues in interval + 1, but then
    !       many restarts are needed. For large values of JMAX, which is the
    !       preferred choice, MUL+1 runs are needed, where MUL is the maximum
    !       multiplicity of an eigenvalue in the interval.
    !         Works on nonsymmetric as well as symmetric pencils, but then accuracy
    !       is approximately TOL*(Henrici departure from normality). The
    !       parameter SPD is used only to choose between SYMMMD and COLMMD when
    !       factorizing, the former being marginally better for symmetric matrices
    !       close to the lower end of the spectrum.
    !        case of trouble:
    !       If convergence is too slow, try (in this order of priority)
    !       1) a smaller interval LB,UB
    !       2) a larger JMAX
    !       3) a larger MAXMUL.
    !         If factorization fails, try again with LB and UB finite.
    !       Then shift is chosen at random and hopefully not at an eigenvalue.
    !       If it fails again, check whether pencil is singular.
    !         If it goes on forever, there may be too many eigenvalues in the
    !       strip. Try with a small value MAXMUL=2 and see which eigenvalues you
    !       get! Those you get are some of the evs, but a negative iresult tells
    !       you that you have not gotten them all.
    !         If memory overflows, try smaller JMAX.
    !         Algorithm is designed for eigenvalues close to the real axis. If you
    !       want those close to the imaginary axis, try A=i*A!
    !         When SPD is true, shift is at LB so that advantage is taken of the
    !       faster factorization for symmetric positive definite matrices.
    !       No harm done but slower execution if LB above lowest eigenvalue.

    !       A Ruhe 8-08-94, AN 8-23-94.
    !       Copyright (c) 1994-98 by The MathWorks, Inc.
    !       $Revision: 1.7 $  $Date: 1997/11/25 17:06:49 $

    !Funcions externes
    real(kind=r_kind), external :: dlamch, dnrm2,ddot

    !Paràmetres d'entrada
    integer, intent(in)           :: n
    real(kind=r_kind), intent(in) :: Toe(:), lb, ub
    integer, intent(in), optional :: jmax_, maxmul_
    real(kind=r_kind), intent(in), optional :: tolconv_

    !Parametres d'eixida
    real(kind=r_kind), intent(inout) :: xv(:,:), lmb(:) 
    integer, intent(inout) :: iresult

    !Variables on copiarem els paràmetres d'entrada opcionals
    integer :: jmax, maxmul

    !Constants del model numeric
    real(kind=r_kind) :: inf, eps 
    
    !Variables auxiliars
    real :: ranv(2), shift !sembla k random_number vol reals de simple precissi. comprovar-ho + avant
    real, allocatable :: rand_n(:)
    integer :: ntries, ierr, i, info, js, jf, ntc, mul
    integer, parameter :: maxtries =2
!    real(kind=r_kind) :: C(n), L(n,n/2 + mod(n,2)), D(n)
    real(kind=r_kind),allocatable :: C(:), C2(:), L(:,:)
    integer, allocatable :: inside(:), ievsort(:)
    real(kind=r_kind) :: fact, tol, tolspur, tolstab, normhj, tolconv, t1, t2, tn
    real(kind=r_kind), allocatable :: v(:,:), h(:,:), T(:,:), US(:,:), a(:,:), wr(:), wi(:), work(:), vs(:,:), vs2(:,:), sij(:), rp(:,:)
!    real(kind=r_kind) :: vstart(n), vj1(n), hj(n+1)
    real(kind=r_kind), allocatable :: vstart(:), vj1(:), hj(:)
!    real(kind=r_kind) :: vec(n), r(n+1)
    real(kind=r_kind), allocatable :: vec(:), r(:)
    integer :: nv, mh, nh, mT, nT, nUS, jt, j, tam, sdim, iintr, nc, k,mrp,nrp,ii
    logical :: convrun, no_use, terminatealg 
    complex(kind=r_kind), allocatable :: lms(:),ev(:), eval(:),evec(:,:)
    logical, allocatable :: convoutside(:), nonconvinside(:), aux_bool(:), newinside(:) 

! #ifdef LEVENIN    
!     ! Parametres del solver Levenin
!     real(kind=r_kind), parameter :: lev_col_tol=0.2, lev_tol=0.3
!     integer, parameter :: k_max=10, refine=0
!     integer, external :: d_lev_inner
! #endif


! #ifdef LEVENIN
!     allocate(C2(2*n-1),stat=ierr) 
!     if (ierr .ne. 0) stop "Error assignant memoria en dtsyeig"
! #else
    allocate(C(n), L(n,n/2 + mod(n,2)), stat=ierr) 
    if (ierr .ne. 0) stop "Error assignant memoria en dtsyeig"
!#endif

    allocate(rand_n(n), vstart(n), vj1(n), hj(n+1), vec(n), r(n+1),stat=ierr) 
    if (ierr .ne. 0) stop "Error assignant memoria a l'entrar"

    inf = huge(inf)
    eps = epsilon(eps)
    
    nv=0
    mh=0
    nh=0
    mT=0
    nT=0
    nUs=0
    j=0
    ntries=0
    iresult=0

    ! tractament d'arguments opcionals
    if (present(tolconv_)) then
       tolconv = tolconv_
    else
       tolconv = 100 * eps
    end if

    if (present(jmax_)) then
       jmax = jmax_
    else
       jmax = min(100,n)
    end if

    if (present(maxmul_)) then
       maxmul = maxmul_
    else
       maxmul = n
    end if

    ! Choose shift
    if (lb .eq. -inf .and. ub .eq. inf) then 
       ! No info on eigenvalue location, pick shift in the dark
       call random_number(shift)
    elseif (lb .eq. -inf) then
       shift = ub
    elseif  (ub .eq. inf) then
       shift = lb
    else
       call random_number(ranv)

       !shift = (lb * ranv(1) + ub * ranv(2)) / sum(ranv)

       !DEGUB
       shift=(lb+ub)/2.d0
    end if

    ! Never more than n basis vectors
    jmax = min(jmax,n)

    ! Factorize and check if problem it not too singular
    do while (ntries .le. maxtries)

       ! Resolucio sistema C*r = rand(n,1)
       call random_number(rand_n)

!call ctimer(t1, tn, tn)
! #ifdef LEVENIN
!        ! C = Toe-shift*I
!        C2(n:) = Toe
!        C2(n) = C2(n) - shift
!        do i=1,n-1
!           C2(n-i) = C2(n+i)
!        end do

!        ! Resolucio sistema C*r = rand(n,1)
!        ierr = d_lev_inner(r, C2(n), n, dble(rand_n), lev_col_tol, lev_tol, k_max, refine)

!        ! Calcul norma
!        fact = maxval(abs(r)) * dtsy_norminf(C2(n:))
! #else
       ! C = Toe-shift*I
       C = Toe
       C(1) = C(1) - shift

       ! Descomposicio LDL^T de C
       !call dtsyldl(n, C, L)
       call dtsyldl(C, L)

       ! Resoluci LDL^T*r = rand(n,1)
       call dtsysv(L, r, dble(rand_n))
       ierr = 1
       !print *,'sol 1r stema',r

       ! Calcul norma
       fact = maxval(abs(r)) * dtsy_norminf(C)
!#endif
!call ctimer(t2, tn, tn)
!print *, 'solve', t2-t1

!       print *, 'norma solucio:', dnrm2(n,r,1), 'shift', shift, t2-t1, 'fact', fact, 0.01/eps, 'ierr',ierr
!       stop

       if (fact .gt. (0.01/eps)) then
          ! Matrix C is singular, try another shift
          if (lb .eq. -inf .and. ub .eq. inf) then
             ! No info on eigenvalues location, pick shift in the dark
             call random_number(shift)
          elseif (lb .eq. -inf) then
             if (ub .eq. 0) then
                ! No info here either
                call random_number(shift)
                shift = -shift
             else
                call random_number(shift)
                shift = ub-abs(ub)*shift
             end if
          elseif (ub .eq. inf) then
             if (lb .eq. 0) then
                ! No info here either
                call random_number(shift)
             else
                call random_number(shift)
                shift = lb+abs(lb)*shift
             end if
          else
             call random_number(ranv)
!             shift = (lb * ranv(1) + ub * ranv(2)) / sum(ranv)
             shift =(lb+ub)/2.d0
          end if
       else
          ! Matrix C is not singular, continue with the algorithm
          exit ! from do while
       end if

       ntries = ntries +1
    end do ! end of do while

    if (ntries .gt. maxtries) then
       print *, 'Suitable eigenvalue shift could not be found. Try a different range.'
       return
    end if

    !DEBUG
    !print *, 'shift',shift

    ! tol is tolerance for convergence
    ! tolspur tol to determine spurious vectors
    ! tolstab decides when eigenvalue appr is stabilized
    tol = tolconv
    tolspur = 100 * tolconv
    tolstab = 1e-5_r_kind 
    ! check for convergence first after jf steps, then every js steps
    ! The larger n the cheaper is the test compared to an extra step
    js = ceiling(4000.0/n)
    jf = 10
    normhj = 0

    ! ntc is total number of converged values, both inside and outside interval
    ! US is n*nT matrix of Schur vectors
    ! T is nT*nT upper triangular Schur form
    ntc=0; 

    ! no hi ha cap reserva abans
!    if (allocated(v)) deallocate(v,stat=ierr)
!    if (ierr .ne. 0) stop 'Error alliberant memoria en misptarn()_0' 

    ! Aci el que fem es una previssio de l'espai final, caldra refinar-la 
    allocate(US(n,jmax),T(jmax,jmax),v(n,jmax+2),h(jmax+1,jmax),stat=ierr) 
    if (ierr .ne. 0) stop 'Error assignant memoria en dtsyeig' 

    ! d'US, T, v i h anem controlant les dimensions amb mUS, nUS, etc. 

    allocate(a(jmax,jmax),wr(jmax),wi(jmax),vs(jmax,jmax),work(3*jmax),lms(jmax),sij(jmax),convoutside(jmax),nonconvinside(jmax),aux_bool(jmax),newinside(jmax),stat=ierr) 
    if (ierr .ne. 0) then 
       print *, 'Error assignant memoria en misptarn()_3' 
       return 
    end if

    ! Starting vector
    call random_number(vstart)
    vstart=1_r_kind
    terminatealg = .false.
    
    ! Finally, it is time to START
    do mul=1,maxmul
       !print *, 'bucle exterior:', mul

       ! Starting vector orthogonalized against already converged
       ! v := [US pdegrmsc(vstart,US)] 
       call pdegrmsc(vstart,US,n,nUS,vj1,hj) 
       v(:,1:nUS) = US(:,1:nUS) 
       v(:,nUS+1) = vj1 
       nv = nUS+1 

       ! h := [T; zeros(1,ntc)] 
       h(1:mT,1:nT) = T(1:mT,1:nT) 
       h(mT+1,1:ntc) = 0 
       if (ntc .gt. 0) mh = mT+1 
       nh = nT 

       !print *, 'v', v(:,:nv)
       !print *, 'h', h(:mh,:nh)

       convrun=.false. 
       !First testing point 
       jt = min(ntc+jf,jmax) 
       do j=ntc+1,jmax           
!          print *,'bucle interior:',j

!call ctimer(t1, tn, tn)
! #ifdef LEVENIN
!           ! Resolucio sistema C*r = v(;,j)       
!           ierr = d_lev_inner(r, C2(n), n, v(1,j), lev_col_tol, lev_tol, k_max, refine)
! #else
          ! Resolucio sistema C*r = v(;,j)
          call dtsysv(L, r, v(:,j))
!#endif
!call ctimer(t2, tn, tn)
!print *, 'solve', t2-t1

!       print *, 'norma solucio:', dnrm2(n,r,1), 'shift', shift, t2-t1
!           print *, 'v despres', v(:,:nv)
!           print *, 'h', h(:mh,:nh)
!           print *, 'r', r(:,1)

          call pdegrmsc(r,v(:,1:nv),n,nv,vj1,hj) 
!        vj1(1:n)=r(1:n)
!        do ii=1,nv
!           hj(ii)=ddot(n,vj1(1:n),1,v(:,ii),1)
!           vj1(1:n)=vj1(1:n)-hj(ii)*v(:,ii)
!        end do
!        hj(nv+1)=dnrm2(n,vj1(1:n),1)
!        if (hj(nv+1).eq.0.0) stop
   !    v(:,nv+1)=vj1(1:n)/hj(nv+1)

          ! h := [h hj(1:j)] 
          h(1:j,nh+1) = hj(1:j) 
          if (mh .eq. 0) mh = 1 
          nh = nh+1 

          normhj = max(normhj, dnrm2(nv+1,hj,1))
!           print *, 'hj:', hj
!           print *, 'normhj:', normhj, dnrm2(nv+1,hj,1)

          if (hj(j+1) .lt. tol*normhj) then 
             convrun = .true. 
             jt = j 
          else 
             ! v := [v vj1] 
             v(:,nv+1) = vj1(1:n) 
             nv = nv+1 
          end if
          ! h := [h;zeros(1,j-1) hj(j+1)] 
          h(mh+1,1:(j-1)) = 0_r_kind !(/(0, i=1,j-1)/) 
          h(mh+1,j) = hj(j+1) 
          mh = mh+1 
          
          ! One vector added, is it time to test? 
          if (j .eq. jt) then 
             tam = jt-ntc
             !print *, 'tam:',tam
             a(:tam,:tam)=h(ntc+1:jt,ntc+1:jt) 

             call dgees('V','N',no_use,tam,a,jmax,sdim,wr,wi,vs,jmax,work,3*jmax,no_use,info) 

             if (info .ne. 0) then 
                print *, 'Error calculat factoritzacio de Schur en misptarn()' 
                return 
             end if

             lms(:tam) = shift + 1/cmplx(wr(:tam),wi(:tam)) 
             sij(:tam) = abs((h(j+1,j)/normhj) * vs(j-ntc,:tam))


             ! Test if it is time to stop 
             convoutside(:tam) = (dble(lms(:tam))<=lb .or. dble(lms(:tam))>ub) .and. sij(:tam)<=tolstab ! aci la funcio dble() extrau la part real del complexe 
             nonconvinside(:tam) = lb<dble(lms(:tam)) .and. dble(lms(:tam))<=ub .and. sij(:tam)>tol .and. abs(vs(1,:tam))>tolspur 
             convrun = convrun .or. (any(convoutside(:tam)) .and. .not.any(nonconvinside(:tam))) 

             ! Report progress if wanted 
             if (any(nonconvinside(:tam))) then 
                do iintr=1,tam 
                   if (nonconvinside(iintr)) exit 
                end do
                ! en este punt iintr conte l'index del primer element a .true. del vector nonconvinside
             else 
                iintr = 1 
             end if
             !            print *, "Aci va la informacio dels autovalors" 
             ! end report part 
             jt = min(jt+js,jmax) 
          end if
          if (convrun) exit 
       end do ! do j=ntc+1,jmax 

       ! nc is number of converged eigenvalues above the first nonconverged along h 
       nc = 0 
       aux_bool(:tam) = sij(:tam) .le. tol 
       do i=1,tam 
          if (.not. aux_bool(i)) exit 
          nc = nc+1 
       end do

       ! Report progress of sweep 
       !       print *, "Aci va la informacio dels autovalors" 
       ! end report progress 

       if (nc .gt. 0) then 
          ! T=[T h(1:ntc,ntc+1:jt)*us(:,1:nc); zeros(nc,ntc) ts(1:nc,1:nc)]; 
          if (ntc .eq. 0) then 
             T(1:nc,1:nc) = a(1:nc,1:nc) 
             nT = nc 
             mT = nc 
          else 
             !T(1:ntc,mT+1:mT+nc) = matmul(h(1:ntc,ntc+1:jt), us(:,1:nc))
             call dgemm('n','n',ntc,nc,tam,dble(1),h(1,ntc+1),jmax+1,us(1,1),n,dble(0),T(1,mT+1),jmax)
             T(ntc+1:ntc+nc,1:ntc) = 0_r_kind 
             T(ntc+1:ntc+nc,mT+1:mT+nc) = a(1:nc,1:nc) 
             mT = mT+nc 
             nT = ntc+nc 
          end if

          !           US=[US v(:,ntc+1:jt)*us(:,1:nc)]; 
          !          US(:,nUS+1:nUS+nc) = matmul(v(:,ntc+1:jt),vs(:,1:nc))
          call dgemm('n','n',n,nc,tam,dble(1),v(1,ntc+1),n,vs(1,1),jmax,dble(0),US(1,nUS+1),n)
          nUS = nUS+nc 

          ntc = ntc+nc 
          if (ntc .ge. jmax) terminatealg = .true. 
       end if
       newinside(:tam) = lb<dble(lms(:tam)) .and. dble(lms(:tam))<=ub 

       if (nc>0 .and. any(newinside(:tam) .and. sij(:tam)<=tol)) then 
          call random_number(vstart) 
       elseif (any(newinside(:tam))) then 
          do i=1,tam 
             if (newinside(i)) exit 
          end do
          !vstart = matmul(v(:,ntc+1:jt), vs(:,i))
          call dgemv('n',n,tam,dble(1),v(1,ntc+1),n,vs(1,i),1,dble(0),vstart,1)
       else 
          terminatealg = .true. 
       end if

       if (terminatealg) exit
    end do ! do mul=1,maxmul 


    if (mT .eq. 0) then
       ! No em trobat autovalors en aquest interval i no cal fer res mes
       iresult = 0
       return
    end if

    ! N'havem trobat d'autovalors	
    ! Triangular matrix is complete. Get eigensolution 
    if (allocated(wr)) deallocate(wr,wi,vs,work)             
    allocate(wr(mT),wi(mT),vs(mT,mT),work(4*mT),stat=ierr) 
    !    if (allocated(eval)) deallocate(eval,evec)             
    !    allocate(eval(mT),evec(mT,mT),rp(mT,mT)) 
    if (ierr .ne. 0) then 
       print *, 'Error assignant memoria en misptarn()_4' 
       return 
    end if
    !      CALL DEVCRH (mT, T, jmax, EVAL, EVEC, mT)

    !	print *, 'Valor de mT: ',mT
    call dgeev('N','V',mT,T,jmax,wr,wi,vs,mT,vs,mT,work,4*mT,info)
    if (info .ne. 0) then 
       print *, 'Error calculat factoritzacio de Schur en misptarn()' 
       return 
    end if
    
    allocate(ev(mT),stat=ierr) 
    if (ierr .ne. 0) then 
       print *, 'Error assignant memoria en misptarn()_5' 
       return 
    end if

    ev = shift + 1/cmplx(wr,wi)

    deallocate(aux_bool);allocate (aux_bool(size(ev))) 
    aux_bool = (lb<dble(ev)) .and. (dble(ev)<=ub)
    iresult = count(aux_bool) 

    ! inside = find(lb<ev & ev<=ub) 
    allocate(inside(iresult),ievsort(iresult)) 
    k=1 
    do i=1,size(ev) 
       if (aux_bool(i)) then 
          inside(k) = i 
          k = k+1 
       end if
    end do

    ! [lmb,ievsort]=sort(ev(inside)) 
!    print *,'aci', iresult, size(lmb,1)
    lmb(1:iresult) = ev(inside) 
!    print *,'aci', iresult
    call mrgrnk(lmb,ievsort) 
    lmb(1:iresult) = lmb(ievsort) 

!    xv(:,1:iresult) = matmul(US(:,1:nUs),vs(:,inside(ievsort)))
    allocate(vs2(mT,iresult))
    vs2=vs(:,inside(ievsort))
    call dgemm('n','n',nUS,iresult,nUS,dble(1),US,n,vs2,mT,dble(0),xv,size(xv,1))

    if (.not. terminatealg) iresult = -iresult 

    deallocate(inside, ievsort, v, h, T, US, a, wr, wi, work, vs, vs2, sij, lms, ev, convoutside, nonconvinside, aux_bool, newinside)
    
    deallocate(rand_n, vstart, vj1, hj, vec, r) 

! #ifdef LEVENIN
!     deallocate(C2)
! #else
    deallocate(C,L)
!#endif

  end subroutine dtsyeig


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


  subroutine dtsyeig_arp (n, Toe, lb, ub, eig_in, xv, lmb, iresult, tolconv_, jmax_, maxitr_)

    !Paràmetres d'entrada
    integer, intent(in)           :: n, eig_in
    real(kind=r_kind), intent(in) :: Toe(:), lb, ub
    integer, intent(in), optional :: jmax_, maxitr_
    real(kind=r_kind), intent(in), optional :: tolconv_

    !Parametres d'eixida
    real(kind=r_kind), intent(inout) :: xv(:,:), lmb(:) 
    integer, intent(inout) :: iresult

    !Variables on copiarem els parametres d'entrada opcionals
    integer :: jmax, maxitr
    real(kind=r_kind) :: tolconv

    ! Variables locals
    real(kind=r_kind) :: shift
    character :: bmat, which*2
    integer :: lworkl, ido, info, ierr, iparam(11), ipntr(11), i
    logical :: rvec

    real(kind=r_kind), allocatable :: C(:), C2(:), L(:,:), resid(:), workd(:), workl(:)
    logical, allocatable :: select(:)

! #ifdef LEVENIN    
!     ! Parametres del solver Levenin
!     real(kind=r_kind), parameter :: lev_col_tol=0.2, lev_tol=0.3
!     integer, parameter :: k_max=10, refine=0
!     external :: d_lev_inner
! #endif


!#ifdef LEVENIN
!    allocate(C2(2*n-1),stat=ierr) 
!    if (ierr .ne. 0) stop "Error assignant memoria en dtsyeig"
!#else
    allocate(C(n), L(n,n/2 + mod(n,2)), stat=ierr) 
    if (ierr .ne. 0) stop "Error assignant memoria en dtsyeig"
!#endif


     ! tractament d'arguments opcionals
    if (present(tolconv_)) then
       tolconv = tolconv_
    else
       tolconv = 100 * epsilon(tolconv)
    end if

    if (present(jmax_)) then
       jmax = jmax_
    else
       jmax = min(100,n)
    end if

    if (present(maxitr_)) then
       maxitr = maxitr_
    else
       maxitr = n
    end if

    ! Reserva de memoria
    lworkl = jmax*(jmax+8)
    allocate(resid(n), workd(3*n), workl(lworkl), select(jmax), stat=ierr) 
    if (ierr .ne. 0) stop "Error assignant memoria en dtsyeig_arp"    

!#ifdef LEVENIN
!    ! C = Toe-shift*I
!    C2(n:) = Toe
!    C2(n) = C2(n) - shift
!    do i=1,n-1
!       C2(n-i) = C2(n+i)
!    end do
!#else
    ! C = Toe-shift*I
    shift=(lb+ub)/2
    C = Toe
    C(1) = C(1) - shift
    
    ! Descomposicio LDL^T de C
    !call dtsyldl(n, C, L)
    call dtsyldl(C, L)
!#endif
    
    ! Parametres ARPACK
    bmat = 'I'
    which = "LM"
    ido = 0
    info = 0
    
    iparam(1) = 1
    iparam(3) = maxitr
    iparam(7) = 3
    
    ! Bucle principal (comunicaci inversa)
    do
       !call dsaupd ( ido, bmat, n, which, eig_in, tolconv,   &
       !     resid, jmax, xv, n, iparam, ipntr, workd, workl, &
       !     lworkl, info )
       if (info .ne. 0) then
          print *, "Error en algorisme de Lanczos", info
          return
       end if
       
       if (ido .eq. -1 .or. ido .eq. 1) then
          ! Resolucio sistema
!#ifdef LEVENIN
!          call d_lev_inner(workd(ipntr(2)), C2(n), n, workd(ipntr(1)), lev_col_tol, lev_tol, k_max, refine)
!#else
          !call dtsysv(n, L, workd(ipntr(2):ipntr(2)+n), workd(ipntr(1):ipntr(1)+n))
 !         print *, ipntr(1), ipntr(2)
  !        print *, 'abans', workd(ipntr(1)), workd(ipntr(1)+1), workd(ipntr(2)), workd(ipntr(2)+1)
!          call dtsysvr(n, L, workd(ipntr(2):ipntr(2)+n), workd(ipntr(1):ipntr(1)+n), C, 0)
!          print *, 'dsprs', workd(ipntr(1)), workd(ipntr(1)+1), workd(ipntr(2)), workd(ipntr(2)+1)
!#endif
       else
          exit
       end if
    end do

!    print *, 'its', iparam(3), 'solves', iparam(9), 'reorto', iparam(11)

    ! No hi ha hagut errors apliquem postprocs
    rvec = .true.
    !call dseupd ( rvec, 'All', select, lmb, xv, n, shift,     &
    !     bmat, n, which, eig_in, tolconv, resid, jmax, xv, n, &
    !     iparam, ipntr, workd, workl, lworkl, ierr )
    if (info .ne. 0) then
       print *, "Error en postproces de Lanczos", info
       return
    end if

    iresult = iparam(5)

  end subroutine dtsyeig_arp

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  subroutine dtsy_levdur(n, T, lambda, q, dq, y)
    ! ..Input/Output Arguments..
    integer, intent(in) :: n
    real(kind=r_kind), intent(in) :: T(:), lambda
    real(kind=r_kind), intent(out) :: q, dq, y(:)

    ! ..Local variables..
    real(kind=r_kind), allocatable :: z(:,:), vec_q(:)
    
    integer info, i, m, j, neg
    real(kind=r_kind) :: aux

    real(kind=r_kind), external :: dnrm2, ddot

    ! ..Executable statements..

    allocate(z(n,n), vec_q(n), stat=info)
    if (info .ne. 0) stop 'Error reservant memoria. Aborte'

    z(1,1)   = T(2)/(T(1) - lambda)
    vec_q(1) = T(1) - lambda

    do m=2,n-1
       vec_q(m) = (1-z(m-1,m-1)**2) * vec_q(m-1)
       z(m,m)   = T(m+1)

       do j=1,m-1
          z(m,m) = z(m,m) - t(m-j+1)*z(j,m-1)
       end do
!       z(m,m) = z(m,m) - dot_product(t(m:2),z(1:m-1,m-1))
!       z(m,m) = z(m,m) - ddot(m-1, t(m), -1, z(1,m-1), 1)
    
       z(m,m)=z(m,m)/vec_q(m);
       
       do j=1,m-1
          z(j,m)=z(j,m-1)-z(m,m)*z(m-j,m-1)
       end do
!       z(1:m-1,m) = z(1:m-1, m-1) - z(m,m)*z(m-1:1,m-1)
!       print *, 'm',m
    end do

    vec_q(n) = (1-z(n-1,n-1)**2)*vec_q(n-1)

    q  = vec_q(n)
    dq = -1.0_r_kind - dnrm2(n,z(1,n-1),1)**2
  
    y(1)  = -1.0_r_kind
    y(2:n) = z(1:n-1,n-1)

  end subroutine dtsy_levdur

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  integer function dtsy_neg_lev(n,T)
    ! ..Input/Output Arguments..
    integer, intent(in) :: n
    real(kind=r_kind), intent(in) :: T(:)

    ! ..Local variables..
    real(kind=r_kind), allocatable :: z(:,:), q(:)
    real(kind=r_kind) :: dq
    
    integer info, i, m, j, neg

    real(kind=r_kind), external :: dnrm2

    ! ..Executable statements..

    allocate(z(n,n), q(n), stat=info)
    if (info .ne. 0) stop 'Error reservant memoria en dtsy_neg_lev. Aborte'

    z(1,1) = T(2)/T(1)
    q(1)   = T(1)

    do m=2,n-1
       q(m)   = (1-z(m-1,m-1)*z(m-1,m-1)) * q(m-1)
       z(m,m) = T(m+1)

       do j=1,m-1
          z(m,m) = z(m,m) - t(m-j+1)*z(j,m-1)
       end do
    
       z(m,m)=z(m,m)/q(m);
       
       do j=1,m-1
          z(j,m)=z(j,m-1)-z(m,m)*z(m-j,m-1);
       end do
    end do

    q(n)=(1-z(n-1,n-1)**2)*q(n-1);

    neg = count(q<0_r_kind)

    dtsy_neg_lev = neg

  end function dtsy_neg_lev

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  integer function flipud_v(v, flip_v)
    real(kind=r_kind), intent(in)  :: v(:)
    real(kind=r_kind), intent(out) :: flip_v(:)

    integer :: i, n
    
    n = size(v)

    if (size(flip_v) .ne. n) then
       flipud_v = -1
       return
    end if

    do i=1,n
       flip_v(i) = v(n-i+1)
    end do

    flipud_v = 0
    
  end function flipud_v


  integer function flipud_m(v, flip_v)
    real(kind=r_kind), intent(in)  :: v(:,:)
    real(kind=r_kind), intent(out) :: flip_v(:,:)

    integer :: i, j, mv ,nv
    
    mv = size(v,1)
    nv = size(v,2)

 !   print *, mv, nv

    if ((mv .ne. size(flip_v,1)) .or. (nv .ne. size(flip_v,2))) then
       flipud_m = -1
       return
    end if

    do j=1,nv
       do i=1,mv
          flip_v(i,j) = v(mv-i+1,j)
       end do
    end do

    flipud_m = 0
    
  end function flipud_m

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  integer function find(v, inds)
    logical, intent(in) :: v(:)
    integer, intent(out) :: inds(:)

    integer :: n,i,k

    if (size(inds) .lt. count(v)) then
       find =-1
       return
    end if

    n = size(v)

    k = 1
    do i=1,n
       if (v(i)) then
          inds(k) = i
          k = k +1
       end if
    end do
    
    find = 0
    
  end function find

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  integer function dtsyeig_nlanc (n, T, lb, ub, n_autov_in, tolconv, autoval, autovec)
    ! Input parameters
    integer, intent(in) :: n, n_autov_in
    real(kind=r_kind), intent(in) :: T(:), lb, ub, tolconv

    ! Ouput parameters
    real(kind=r_kind), intent(out) :: autoval(:), autovec(:,:)
    
    ! Local variables
    logical, allocatable, dimension(:) :: inda_inner, inds_inner, convs_inside, conva_inside
    integer :: m, ierr, itmax, maxmul, n_convs, n_conva, i, r, reorts, npg, nqg, k, j, lwork, liwork, &
         n_lms, n_lma, n_inds, n_inda, n_vs_conv, n_va_conv, n_inds_inner, n_inda_inner
    integer, allocatable, dimension(:) :: iwork, isuppz, inds, inda
    real(kind=r_kind) :: shift, reps, nh1s, nh1a
    real(kind=r_kind), allocatable, dimension(:)   :: C, p, q, w, h1s, h1a, autoval_convs, autoval_conva, &
         v, vs, va, ms_diag, ma_diag, ms_subdiag, ma_subdiag, p_old, q_old, work, lms, lma, errs, erra, &
         aux_m
    real(kind=r_kind), allocatable, dimension(:,:) :: L, Pg, Qg, autovec_convs, autovec_conva, &
         vs_conv, va_conv, evs, eva, auxvs, auxva, autos, autoa
    

    ! External references
    real(kind=r_kind), external :: ddot, dnrm2
    external :: dstevr

    real( kind=r_kind ) :: t1, t2, tn

    !
    ! Input parameters checking    
    !
    if (mod(n,2) .ne. 0) then
       print *, 'Only problems with even n can be solved!'
       dtsyeig_nlanc = -1
       return
    end if

    if (size(T) .ne. n) then
       print *, 'Array T must be of size n!'
       dtsyeig_nlanc = -1
       return
    end if

    if ((size(autoval) .lt. n_autov_in) .or. (size(autovec) .lt. n*n_autov_in)) then
       print *, 'Arrays autoval and autovec must be, at least, of size n_autov_in and n*n_autov_in!'
       dtsyeig_nlanc = -1
       return
    end if

    !
    ! Auxiliar variables inicialisation
    !
    ! m is the size of each subproblem (symmetric and antisymmetric)
    m = n/2
    ! max number of iterations of each Lanczos run
    itmax = 30 + 2*n_autov_in
    ! max number of Lanczos runs (restarts)
    maxmul = 4
    ! number of reortogonalisations per restart
    reorts = 2
    ! number of already converged eigenvalues(vectors)
    n_convs = 0
    n_conva = 0
    ! # of columns in each basis (symmetric and antisymmetric)
    npg = 0
    nqg = 0   
    !
    reps = 10 * sqrt(epsilon(reps))
    ! workspace required by LAPACK dstver()
    lwork = 20*n 
    liwork = 10*n
    ! # of columns in vs_conv and va_conv
    n_vs_conv = 0
    n_va_conv = 0
    !
    n_inds_inner = 0
    n_inda_inner = 0


    !
    ! Memory allocation
    !
    allocate(C(n), L(n,n/2 + mod(n,2)), Pg(m, itmax), Qg(m, itmax), p(m), q(m), w(n), &
         h1s(itmax), h1a(itmax), v(n), vs(m), va(m), ms_diag(itmax), ma_diag(itmax),&
         ms_subdiag(itmax), ma_subdiag(itmax), p_old(m), q_old(m), work(lwork), iwork(liwork), stat=ierr)
    if (ierr .ne. 0) then
       print *, 'Memory allocation error dtsyeig_nlanc 1!'
       dtsyeig_nlanc = -1
       return
    end if

    allocate(isuppz(2*n), lms(itmax), lma(itmax), evs(itmax,itmax), eva(itmax,itmax), errs(itmax), &
         erra(itmax), convs_inside(itmax), conva_inside(itmax), aux_m(m), vs_conv(m,itmax), &
         va_conv(m,itmax), &
    !    autoval_convs(itmax), autoval_conva(itmax), autovec_convs(n,itmax), autovec_conva(n,itmax), &
         inda_inner(m), inds_inner(m), &
    !     auxvs(n,n_autov_in), auxva(n,n_autov_in),&
         autos(n,n_autov_in), autoa(n,n_autov_in), stat=ierr) 
    if (ierr .ne. 0) then
       print *, 'Memory allocation error dtsyeig_nlanc 2!'
       dtsyeig_nlanc = -1
       return
    end if

    !
    ! Beggining of the algorithm
    !   
     
    ! The middle point of the subinterval is used as shift
    shift = (lb+ub) / 2
   
    ! LDLT factorisation of the shifted Toeplitz matrix
    C = T
    C(1) = C(1) - shift
    !print *,'C = ',C
    !call dtsyldl(n, C, L)
    call dtsyldl(C, L)
    !print *,'shift = ',shift

    ! Initial arrays for the recursion
    call random_number(p)
    call random_number(q)
    p = p / sqrt(2*ddot(m, p, 1, p, 1))
    q = q / sqrt(2*ddot(m, q, 1, q, 1))

    ! Lanczos outter loop (restarts of the algorithm)
    do i=1,maxmul

       if (n_convs .gt. 0) then
          ! In the case of restart, vs_conv stores the eigenvalues of the symmetric subproblem
          ! already converged, those eigenvalues are reortogonalised and used as the new Lanczos Pg basis

! Posar amb un if com a la reortonormalitzacio

          do r=1,reorts
             ! h1s = vs_conv' * p
             call dgemv('T', m, n_vs_conv,  1.0_r_kind, vs_conv, m, p, 1, 0.0_r_kind, h1s, 1)
             ! p = p - 2*vs_conv*h1s
             call dgemv('N', m, n_vs_conv, -2.0_r_kind, vs_conv, m, h1s, 1, 1.0_r_kind, p, 1)
          end do
          p = p / 2*ddot(m, p, 1, p, 1)

          Pg(:,:n_vs_conv) = vs_conv(:,:n_vs_conv)
          npg = n_vs_conv
       end if

       if (n_conva .gt. 0) then
          ! In the case of restart, va_conv stores the eigenvalues of the antisymmetric subproblem
          ! already converged, those eigenvalues are reortogonalised and used as the new Lanczos Qg basis

! Posar amb un if com a la reortonormalitzacio

          do r=1,reorts
             ! h1a = va_conv' * q
             call dgemv('T', m, n_va_conv,  1.0_r_kind, va_conv, m, q, 1, 0.0_r_kind, h1a, 1)
             ! q = q - 2*va_conv*h1a
             call dgemv('N', m, n_va_conv, -2.0_r_kind, va_conv, m, h1a, 1, 1.0_r_kind, q, 1)
          end do
          q = q / 2*ddot(m, q, 1, q, 1)

          Qg(:,:n_va_conv) = va_conv(:,:n_va_conv)
          nqg = n_va_conv
       end if

       ! Lanczos inner loop (Lanczos algorithm itself)
       do k=1,itmax
          ! A new vector is added to each basis
          Pg(:,npg+1) = p
          npg = npg + 1

          Qg(:,nqg+1) = q
          nqg = nqg + 1

          ! Compound rhs of the linear system from p and q
          w(:m) = p+q
          ierr = flipud_v(p-q, w(m+1:))
          if (ierr .ne. 0) then
             print *, 'Error in flipud_v'
             dtsyeig_nlanc = -2
             return
          end if

!call ctimer( t1, tn, tn )
!print *,'t1'
          ! Tv = w solution via previous LDLT factorisation
          call dtsysv(L, v, w)
          !call dtsysv(L, v, w, C, 1 )
!call ctimer( t2, tn, tn )
!print *,'t2'
tiempo_eigen_toep = tiempo_eigen_toep + (t2-t1)

          ! Obtain symmetric and antisymmetric solution from v
          ierr = flipud_v(v(m+1:), aux_m)
          if (ierr .ne. 0) then
             print *, 'Error in flipud_v'
             dtsyeig_nlanc = -2
             return
          end if
          vs = 0.5_r_kind * (v(1:m) + aux_m)
          va = v(1:m) - vs

          ! Compute diagonals of both tridiagonal matrices (sym and antisym)
          ms_diag(k) = 2.0_r_kind * ddot(m, vs, 1, p, 1)
          ma_diag(k) = 2.0_r_kind * ddot(m, va, 1, q, 1)

          ! vs and va actualisation
          vs = vs-ms_diag(k)*p
          va = va-ma_diag(k)*q

          if (k .gt. 1) then
             vs = vs - ms_subdiag(k-1)*p_old
             va = va - ma_subdiag(k-1)*q_old
          end if
          
          ! Compute subdiagonal elements
          ! h1s = Pg' * vs
          call dgemv('T', m, npg, 1.0_r_kind, Pg, m, vs, 1, 0.0_r_kind, h1s, 1)
          
          ! h1a = Qg' * va
          call dgemv('T', m, nqg, 1.0_r_kind, Qg, m, va, 1, 0.0_r_kind, h1a, 1)

          nh1s = dnrm2(npg, h1s, 1)
          nh1a = dnrm2(nqg, h1a, 1)

          ! vs = vs - 2*Pg*h1s
          call dgemv('N', m, npg, -2.0_r_kind, Pg, m, h1s, 1, 1.0_r_kind, vs, 1)

          ! va = va - 2*Qg*h1a
          call dgemv('N', m, nqg, -2.0_r_kind, Qg, m, h1a, 1, 1.0_r_kind, va, 1)
          
          if (nh1s .gt. reps) then
             ! h1s = Pg' * vs
             call dgemv('T', m, npg, 1.0_r_kind, Pg, m, vs, 1, 0.0_r_kind, h1s, 1)
             ! vs = vs - 2*Pg*h1s
             call dgemv('N', m, npg, -2.0_r_kind, Pg, m, h1s, 1, 1.0_r_kind, vs, 1)
          end if
          

          if (nh1a .gt. reps) then
             ! h1a = Qg' * va
             call dgemv('T', m, npg, 1.0_r_kind, Qg, m, va, 1, 0.0_r_kind, h1a, 1)
             ! va = va - 2*Qg*h1a
             call dgemv('N', m, nqg, -2.0_r_kind, Qg, m, h1a, 1, 1.0_r_kind, va, 1)
          end if

          ms_subdiag(k) = sqrt(2.0_r_kind * ddot(m, vs, 1, vs, 1))
          ma_subdiag(k) = sqrt(2.0_r_kind * ddot(m, va, 1, va, 1))
          ! End computing subdiagonal elements

          ! Prepare next Lanczos iteration
          p_old = p
          q_old = q
          p = vs / ms_subdiag(k)
          q = va / ma_subdiag(k)

          ! Every 5 iterations, starting from iteration number 10, compute ms and ma eigenvalues,
          ! check how many have converged and if they have converged inside the subinterval
          if (((mod(k,5) .eq. 0) .and. (k .ge. 10)) .or. (k .eq. itmax)) then
             ! ms eigenvalues and vectors
             call dstevr('V', 'A', k, ms_diag, ms_subdiag, 0.0_r_kind, 0.0_r_kind, 0, 0, tolconv*0.01, &
                  n_lms, lms, evs, itmax, isuppz, work, lwork, iwork, liwork, ierr)
             if (ierr .ne. 0) then
                print *, 'Error solving symmetric eigenvalue subproblem!', ierr
                dtsyeig_nlanc = -2
                return
             end if
             
             ! ma eigenvalues and vectors
             call dstevr('V', 'A', k, ma_diag, ma_subdiag, 0.0_r_kind, 0.0_r_kind, 0, 0, tolconv*0.01, &
                  n_lma, lma, eva, itmax, isuppz, work, lwork, iwork, liwork, ierr)
             if (ierr .ne. 0) then
                print *, 'Error solving antisymmetric eigenvalue subproblem!', ierr
                dtsyeig_nlanc = -2
                return
             end if

             ! Error estimation
             errs(:n_lms) = abs(ms_subdiag(k)*evs(k,:n_lms)) / maxval(abs(lms(:n_lms)))
             erra(:n_lma) = abs(ma_subdiag(k)*eva(k,:n_lma)) / maxval(abs(lma(:n_lma)))
             
             ! Shift reverse
             lms(:n_lms) = shift + 1/lms(:n_lms)
             lma(:n_lma) = shift + 1/lma(:n_lma)

             ! Boolean array that identifies which eigenvalues have converged inside
             convs_inside(:n_lms) = (lb .le. lms(:n_lms)) .and. (ub .gt. lms(:n_lms)) .and. (errs(:n_lms) .lt. tolconv)
             conva_inside(:n_lma) = (lb .le. lma(:n_lma)) .and. (ub .gt. lma(:n_lma)) .and. (erra(:n_lma) .lt. tolconv)

             if ((count(convs_inside(:n_lms)) + count(conva_inside(:n_lma)) + n_convs + n_conva) .ge. n_autov_in) then
                ! All the eigenvalues in [lb.ub] have converged, prepare exit
                n_inds = count(convs_inside(:n_lms))
                n_inda = count(conva_inside(:n_lma))
                allocate(inds(n_inds), inda(n_inda), stat=ierr) 
                if (ierr .ne. 0) then
                   print *, 'Memory allocation error!'
                   dtsyeig_nlanc = -1
                   return
                end if
                
                ierr = find(convs_inside(:n_lms), inds)
                if (ierr .ne. 0) then
                   print *, 'Error in find!'
                   dtsyeig_nlanc = -2
                   return
                end if

                ierr = find(conva_inside(:n_lma), inda)
                if (ierr .ne. 0) then
                   print *, 'Error in find!'
                   dtsyeig_nlanc = -2
                   return
                end if

! Canviar la linia quan programe el restart
!                autoval = (/lms(inds), eigval_s_conv(:n_eigval_s_conv), lma(inda), eigval_a_conv(:n_eigval_a_conv)/)
                !print *, 'suma', n_inds+n_inda
                autoval(:n_inds+n_inda) = (/lms(inds), lma(inda)/)
!                print *, 'dsprs'

                !
                ! Computing the eigenvectors of the original problem
                !
                ! auxs = Pg(:,nvs+1:nvs+k)*evs(:,inds); autos=[auxs;flipud(auxs)];
                call dgemm('N','N', m, n_inds, k, 1.0_r_kind, Pg(1, n_vs_conv+1), m, evs(:,inds), itmax, 0.0_r_kind, autos, n)
                ierr = flipud_m(autos(:m,:n_inds), autos(m+1:,:n_inds))
                if (ierr .ne. 0) then
                   print *, 'Error in flipud_m'
                   dtsyeig_nlanc = -2
                   return
                end if

                ! auxvs = [Vs_conv(:,inds_inner); flipud(Vs_conv(:,inds_inner))]
                if (n_inds_inner .gt. 0) then
                   auxvs(:m,:n_inds_inner) = vs_conv(:,inds_inner)
                   ierr = flipud_m(auxvs(:m,:n_inds_inner), auxvs(m+1:,:n_inds_inner))
                   if (ierr .ne. 0) then
                      print *, 'Error in flipud_m'
                      dtsyeig_nlanc = -2
                      return
                   end if
                end if

                ! auxa=Qg(:,nva+1:nva+k)*eva(:,inda); autoa=[auxa;-flipud(auxa)];
                call dgemm('N', 'N', m, n_inda, k, 1.0_r_kind, Qg(1,n_va_conv+1), m, eva(:,inda), itmax, 0.0_r_kind, autoa, n)
                ierr = flipud_m(-autoa(:m,:n_inda), autoa(m+1:,:n_inda))
                if (ierr .ne. 0) then
                   print *, 'Error in flipud_m'
                   dtsyeig_nlanc = -2
                   return
                end if

                ! auxva=[Va_conv(:,inda_inner);-flipud(Va_conv(:,inda_inner))];
                if (n_inda_inner .gt. 0) then
                   auxva(:m,:n_inda_inner) = va_conv(:,inda_inner)
                   ierr = flipud_m(-auxva(:m,:n_inda_inner), auxva(m+1:,:n_inda_inner))
                   if (ierr .ne. 0) then
                      print *, 'Error in flipud_m'
                      dtsyeig_nlanc = -2
                      return
                   end if
                end if

                ! autovec = [autos,auxvs,autoa,auxva]
                autovec(:,:n_inds) = autos(:,:n_inds)
                autovec(:,n_inds+1:n_inds+n_inda) = autoa(:,:n_inda)

                dtsyeig_nlanc = (count(convs_inside(:n_lms)) + count(conva_inside(:n_lma)) + n_convs + n_conva)
                
                deallocate(C, L, Pg, Qg, p, q, w, h1s, h1a, v, vs, va, ms_diag, ma_diag, ms_subdiag, ma_subdiag, &
                     p_old, q_old, work, iwork, isuppz, lms, lma, evs, eva, errs, erra, convs_inside, conva_inside, &
                     aux_m, vs_conv, va_conv, &
                     ! autoval_convs, autoval_conva, autovec_convs, autovec_conva, &
                     inda_inner, inds_inner, &
                     ! auxvs, auxva, 
                     autos, autoa, inds, inda, stat=ierr) 
                if (ierr .ne. 0) then
                   print *, 'Memory deallocation error!'
                   dtsyeig_nlanc = -1
                   return
                end if
                
                return
                
             end if
             
             if (k .eq. itmax) then
                print *, 'Restart needed!!!'
                dtsyeig_nlanc = -3
                return
             end if
             
          end if
          
       end do !Lanczos inner loop
       
    end do !Lanczos outter loop

  end function dtsyeig_nlanc

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  function eigen_toep_time
    real(kind=r_kind) :: eigen_toep_time

    eigen_toep_time=tiempo_eigen_toep
  end function eigen_toep_time
  
END MODULE eigen_toep

