#include <time.h>
#include <stdio.h>
#include <stdlib.h>

#include "../../f2c.h"
#define T float
#define Tsf "%f"

typedef struct interval_node_struct {
	double lb, ub;
	integer eigs, prev_eigs, post_eigs;
	struct interval_node_struct *next;
} interval_node;

void calc_intervals(interval_node *ilist, integer *n, double *t, double *lbg, double *ubg, integer *maxEig_interval);

void readInput(int *n, T **v, T *a, T *b, int *nEig) {
	
	// Read matrix size
	scanf("%d", n);

	// Read toeplitz vector
	int i;
	*v = (T *) malloc(*n * sizeof(T));
	for(i=0; i<*n; ++i) scanf(Tsf, (*v) + i);
	
	// Read interval to evaluate
	scanf(Tsf, a);
	scanf(Tsf, b);

	// Read number of eigenvalues
	scanf("%d", nEig);
}

int main(int argc, char *argv[]) {

	// Read input parameters
	int n, nEig;
	T *v = NULL;
	T a, b;
	readInput(&n, &v, &a, &b, &nEig);

	// Convert input vector to double precision
	int i;
	double *t_double = (double *) malloc(n * sizeof(double));
	for(i=0; i<n; ++i) t_double[i] = v[i];

	// Calculate Gershgorin disc interval
	double ubg = 0.0;
	for(i=0; i<n; ++i) ubg += t_double[i];
	double lbg = 2.0 * t_double[0] - ubg;

	// Calculate parameters
	integer n_ = n;
	integer maxEig_interval = 2;
	interval_node intervals;
	intervals.next = NULL;
	calc_intervals(&intervals, &n_, t_double, &lbg, &ubg, &maxEig_interval);


	a = intervals.lb;
	b = intervals.ub;
	nEig = intervals.eigs;
	fprintf(stderr, "Trying to extract %d eigenvalues from interval [%.6f, %.6f]...\n", nEig, a, b);

	// Set tolerance threshold
	/*T tolerance;
	#ifdef SINGLE
	tolerance = 1e-3;
	#else
	tolerance = 1e-14;
	#endif

	clock_t t1 = clock();

	// Shift-and-invert 2-way Lanczos
	T *out_l = (T *) malloc(nEig * sizeof(T));
	T *out_v = (T *) malloc(nEig * n * sizeof(T));
	int extracted = si2w_cpu(n, v, a, b, nEig, 5 * nEig, tolerance, out_l, out_v);

	clock_t t2 = clock();

	// Show results
	fprintf(stderr, "Extracted %d of %d eigenvalues (%.3f)\n", extracted, nEig, (t2 - t1) / (double)CLOCKS_PER_SEC);
	print_vector("eigenvalues", out_l, extracted);
	fprintf(stderr, "\n");
	//print_matrix("eigenvectors", out_v, n, extracted, n);*/
	
	// Free memory
	free(v);
	//free(out_l);
	//free(out_v);

	free(t_double);

	return 0;
}

