module ll_intervals
  implicit none

  integer, parameter, private :: r_kind = kind(1d+0)

  !public :: ll_node, ll_list

  ! Esta constant cal que siga coherent amb el nombre de 
  !camps reals i sencers de la estructura ll_node
  integer, parameter, private :: n_camps = 5

  type ll_node
     real(kind=r_kind) :: lb, ub
     integer :: eigs, prev_eigs, post_eigs
     type(ll_node), pointer :: next => null()
  end type ll_node

  type ll_list
     type(ll_node) :: head
  end type ll_list
  
contains
  
  subroutine ll_init(ll, lb, ub, eigs, prev_, post_)
    type(ll_list), intent(inout) :: ll
    real(kind=r_kind) :: lb, ub
    integer, intent(in) :: eigs
    integer, intent(in), optional :: prev_, post_
    integer :: prev, post
    
    if (present(prev_)) then
       prev = prev_
    else
       prev = 0
    end if

    if (present(post_)) then 
       post = post_
    else
       post = 0
    end if

    allocate(ll%head%next)
    ll%head%next%lb = lb
    ll%head%next%ub = ub
    ll%head%next%eigs = eigs
    ll%head%next%prev_eigs = prev
    ll%head%next%post_eigs = post
!    nullify(ll%head%next%next)
    ll%head%next%next => null()

  end subroutine ll_init

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!1

  integer function ll_length(ll)
    type(ll_list), intent(in), target :: ll
    type(ll_node), pointer :: p_node
    integer :: n

    p_node => ll%head
    n=0
    do while (associated(p_node%next))
       p_node => p_node%next
       n = n+1
    end do

    ll_length = n

  end function ll_length

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!1

  integer function ll_i_node(ll, i, lb, ub, eigs)
    
    type(ll_list), intent(in), target :: ll
    integer, intent(in) :: i
    real(kind=r_kind), intent(out) :: lb, ub
    integer, intent(out) :: eigs

    type(ll_node), pointer :: p_node
    integer :: j
    
    if ((i .lt. 0) .or. (i .gt. ll_length(ll))) then    
       lb=0; ub=0; eigs=0
       ll_i_node = -1
       return
    end if

    p_node => ll%head
    do j=1,i
       p_node => p_node%next
    end do
    
    lb = p_node%lb
    ub = p_node%ub
    eigs  = p_node%eigs
    ll_i_node = i

  end function ll_i_node

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!1

  integer function ll_split_node(ll, i, neg, pos)
    type(ll_list), intent(inout), target :: ll
    integer, intent(in) :: i, neg, pos

    type(ll_node), pointer :: p_node, p_ant
    type(ll_node), pointer :: node1, node2
    integer :: j, n1_eigs, n2_eigs


    if (ll_length(ll) .eq. 0) then
       ll_split_node = -1
       return
    end if

    if ((i .lt. 0) .or. (i .gt. ll_length(ll))) then    
       ll_split_node = -2
       return
    end if

           
    p_ant  => ll%head
    p_node => ll%head%next
    do j=1,(i-1)
       p_ant => p_node
       p_node => p_node%next
    end do

    n1_eigs = neg - p_node%prev_eigs
    n2_eigs = pos - p_node%post_eigs

    if ((n1_eigs .eq. 0) .and. (n2_eigs .eq. 0)) then
       p_ant%next => p_node%next
       deallocate(p_node)

       ll_split_node = 0
       return
    end if

    
    if ((n1_eigs .gt. 0) .and. (n2_eigs .gt. 0)) then
       ! cal afegir els dos fills
       allocate(node1,node2)
       
       p_ant%next => node1
       
       node1%lb   = p_node%lb
       node1%ub   = (p_node%lb + p_node%ub) / 2
       node1%eigs = n1_eigs
       node1%prev_eigs = p_node%prev_eigs
       node1%next => node2
       
       node2%lb   = (p_node%lb + p_node%ub) / 2
       node2%ub   = p_node%ub
       node2%eigs = n2_eigs
       node2%post_eigs = p_node%post_eigs
       node2%next => p_node%next
       
       node1%post_eigs = p_node%post_eigs + n2_eigs
       node2%prev_eigs = p_node%prev_eigs + n1_eigs
              
       ll_split_node=2

    elseif (n1_eigs .gt. 0) then
       ! cal afegir el fill esquerre
       allocate(node1)

       p_ant%next => node1
       
       node1%lb   = p_node%lb
       node1%ub   = (p_node%lb + p_node%ub) / 2
       node1%eigs = n1_eigs
       node1%prev_eigs = p_node%prev_eigs
       node1%post_eigs = p_node%post_eigs

       node1%next => p_node%next

       ll_split_node=1
       
    elseif (n2_eigs .gt. 0) then
       !cal afegir el fill dret
       allocate(node2)

       p_ant%next => node2

       node2%lb   = (p_node%lb + p_node%ub) / 2
       node2%ub   = p_node%ub
       node2%eigs = n2_eigs
       node2%post_eigs = p_node%post_eigs
       node2%prev_eigs = p_node%prev_eigs

       node2%next => p_node%next

       ll_split_node=1

    end if
    
    deallocate(p_node)

  end function ll_split_node

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!1

  subroutine ll_remove_node(ll, i)
    type(ll_list), intent(inout), target :: ll
    integer, intent(in) :: i

    type(ll_node), pointer :: p_node, p_ant
    integer :: j
    type(ll_node), pointer :: node1, node2
    
    p_ant  => ll%head
    p_node => ll%head%next
    do j=1,(i-1)
       p_ant => p_node
       p_node => p_node%next
    end do

    p_ant%next => p_node%next
    deallocate(p_node)

  end subroutine ll_remove_node

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  subroutine ll_finalize(ll)
    type(ll_list), intent(inout), target :: ll
    integer :: i, n

    n = ll_length(ll)

    do i=1, n
       call ll_remove_node(ll,1)
    end do

  end subroutine ll_finalize

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  integer function ll_serialize_mem(ll, ini_, fi_)
    type(ll_list), intent(in), target :: ll
    integer, intent(in), optional :: ini_, fi_

    integer :: ini, fi

    if (present(ini_)) then
       ini = ini_
    else
       ini = 1
    end if
    
    if (present(fi_)) then
       fi = fi_
    else
       fi = ll_length(ll)
    end if


    ll_serialize_mem = (fi-ini+1) * n_camps

  end function ll_serialize_mem

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  subroutine ll_serialize(ll, ll_array, ini_, fi_)
    type(ll_list), intent(in), target :: ll
    real(kind=r_kind), intent(out) :: ll_array(:)
    integer, intent(in), optional :: ini_, fi_
 
    type(ll_node), pointer :: p_node
    integer :: i, j, n, ini, fi
    integer :: id_proc, ierror

    if (present(ini_)) then
       ini = ini_
    else
       ini = 1
    end if

    if (present(fi_)) then
       fi = fi_
    else
       fi = ll_length(ll)
    end if

    p_node => ll%head
    do j=1,ini-1
       p_node => p_node%next
    end do

    i=1
    do j=ini,fi
       p_node => p_node%next
       ll_array(i)   = p_node%lb
       ll_array(i+1) = p_node%ub
       ll_array(i+2) = real(p_node%eigs, r_kind) 
       ll_array(i+3) = real(p_node%prev_eigs, r_kind)
       ll_array(i+4) = real(p_node%post_eigs, r_kind)
       i = i+5
    end do

  end subroutine ll_serialize

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  integer function ll_add_serialized(ll, n_serial, serial)
    type(ll_list), intent(inout), target :: ll
    integer, intent(in) :: n_serial
    real(r_kind), intent(in) :: serial(:)

    type(ll_node), pointer :: p_node, node
    integer :: i, j

    if (mod(n_serial, n_camps) .ne. 0) then
       ll_add_serialized = -1
       return
    end if

    p_node => ll%head
    do while (associated(p_node%next))
       p_node => p_node%next
    end do

    j = 1
    do i=1, n_serial/n_camps
       allocate(node)
       node%lb   = serial(j)
       node%ub   = serial(j+1)
       node%eigs = int(serial(j+2))
       node%prev_eigs = int(serial(j+3))
       node%post_eigs = int(serial(j+4))
       p_node%next => node
       p_node => node
       j = j+5
    end do

    p_node%next => null()

    ll_add_serialized = 0

  end function ll_add_serialized
    
  
end module ll_intervals




 


