program sequ_nlan
  use precision
  use threadsnumber
  use ioutils
  use dst_lib
  use eigen_toep
  use ll_intervals
  use balance
  use omp_lib
  use cl_args
  use cauchy_inertia
  use cauchy
  use toeplitz_cauchy
  implicit none

  !include 'mpif.h' 

  integer :: n, max_eig
  integer :: natVar

  real(kind=r_kind), allocatable :: T(:), autovalors(:), autovectors(:,:), tmp_vectors(:,:), tmp_valors(:)
  real(kind=r_kind) :: lbg, ubg, lb, ub, amplaria, n_intervals, aux
  integer :: i, j, k, n_autov, tmp_n_autov, status, n_eigs
  real(kind=r_kind) :: t1, t2, tn, tini, tfin, tbfi, tbini
  type(ll_list) :: ll_interv
  
  
  status=cl_get('-n',n) + cl_get('-m',max_eig)
  if (status .lt. 0) then
     print *, 'Us: ./exe -n <talla> -m <grandaria_subintervals> [-l <cota_inferior_espectre>, -u <cota_superior_espectre>]'
     stop
  end if
  
  allocate(T(n), autovalors(n), tmp_vectors(n, max_eig), tmp_valors(max_eig), stat=status) 
  if (status .ne. 0) then
     print *, 'Memory allocation error!'
     stop
  end if
  
  
  ! Obtencin del nmero de threads por variable de entorno si sta existe
  ! Este valor se pone en la variable num_threads del modulo threadsnumber
  natVar = read_nat_env_var( Num_threads_variable )
  if( natVar.ge.0 ) num_threads = natVar

  print *, 'grandaria matriu:', n , 'threads:', num_threads
  print *, 'grandaria subintervals:', max_eig
  print *

  ! Generem una matriu Toeplitz simètrica aleatoria
  call random_seed(put=(/33/))
  call random_number(T)
  
  ! Imprimim matriu
!   print *, 'T=['
!   do i=1,n
!           print *, T(i)
!   end do
!   print *, ']'

!  print *, 'norma(T,inf)=', dtsy_norminf(T)
  
  ! Calculem l'interval on s'ubiquen els autovalors mitjançant cercles de Gershgorin
  call dtsy_eigen_bounds(T,lbg,ubg)

  ! Comprovem si per lnia de comando ens han especificat un subinterval
  status=cl_get('-l',aux)
  if (status .eq. 0) lbg = aux
  status=cl_get('-u',aux)
  if (status .eq. 0) ubg = aux

  print *,'lbg: ',lbg,'ubg: ',ubg
  print *

  ! Etapa de balanceig de la càrrega
  !call ctimer(tbini, tn, tn)
  status = seq_balance(ll_interv, n, T, lbg, ubg, max_eig)
  !call ctimer(tbfi, tn, tn)
  if (status .ne. 0) stop 'Error en balancenjament. Aborte'

  !print *, 'etapa de balancejament:',tbfi-tbini,'segons'
  !stop


  n_intervals = ll_length(ll_interv)

  do i=1, n_intervals
     status = ll_i_node(ll_interv, i, lb, ub, n_eigs)
     !print *,'estat',i
     !print *,'lb:', lb
     !print *,'ub:', ub
     !print *,'n_eigs:', n_eigs
     !print *
  end do

  !print *, 'etapa de balancejament:',tbfi-tbini,'segons'


  ! Etapa de resolució
  n_autov = 0
  !call ctimer(tini,tn,tn)

  do i=1, n_intervals
     status = ll_i_node(ll_interv, i, lb, ub, n_eigs)    
     
     !call ctimer(t1, tn, tn)
     tmp_n_autov = dtsyeig_nlanc(n, T, lb, ub, n_eigs,1e-7_r_kind, tmp_valors, tmp_vectors)
     !call ctimer(t2, tn, tn)
     
     if (tmp_n_autov .gt. 0) then
        autovalors (n_autov+1:n_autov+tmp_n_autov) = tmp_valors(:tmp_n_autov)
!        autovectors(:,n_autov+1:n_autov+tmp_n_autov) = tmp_vectors(:,:tmp_n_autov)
        n_autov = n_autov + tmp_n_autov
     end if
     
     !     print *, '[',lb,',',ub,']:', tmp_n_autov, 'autovalors', ' en', t2-t1
     !     print *, 'long ',ub-lb, 'num ', tmp_n_autov, 'temps ', t2-t1
     !     print *, tmp_valors(:tmp_n_autov)
  end do

  !call ctimer(tfin,tn,tn)
  
  print *
  print *, 'autovalors(1):', autovalors(1)
  print *, 'autovalors(n):', autovalors(n)
  !print *, 'autovalors:', autovalors
  !print *, 'autovectors:', autovectors
  print *, '# autovalors:', n_autov
  print *, 'subintervals generats : ', ll_length(ll_interv)
  !print *
  print *, "temps d'execucio      : ", (tbfi-tbini) + (tfin-tini),'segons'
  print *, 'etapa de balancejament: ', tbfi-tbini,'segons'
  print *, "etapa d'extracci     : ", tfin-tini,'segons'
  print *, 'temps dst:', dst_time()
  print *, 'tipo dst:', dst_type()
  print *, 'tiempo de inertia:',inertia_time(),'segons'
  print *, 'tiempo de cauchy:',cauchy_time(),'segons'
  print *, 'tiempo de eigen_toep:',eigen_toep_time(),'segons'
  print *, 'tiempo de dtsysv:',dtsysv_time(),'segons'
  
  call ll_finalize(ll_interv)
  
end program sequ_nlan
