/***************************************************************************
 *   Copyright (C) 2008 by Leandro Graciá Gil                              *
 *   leagragi@fiv.upv.es                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/**
 *  \file     utils.c
 *  \brief    Utility functions for eigenvalue extraction tool
 *  \author   Leandro Graciá Gil, leagragi@inf.upv.es
 *  \date     18/11/08
*/

// Main project header
#include "toeplitz.h"

/**
 * Auxiliar function for reading input data from standard input.
 *
 * \param n Pointer to variable that will store input data size.
 * \param t Pointer to variable that will store input data vector.
 * \param file File where contents should be read from.
 * \return \c 1 if successful, \c 0 otherwise.
*/
int readInput(int *n, T **t, FILE *file) {

	// Read matrix size
	if(fscanf(file, "%d", n) != 1) return 0;

	// Check size
	if(n <= 0) return 0;

	// Read Toeplitz vector
	int i;
	*t = (T *) malloc(*n * sizeof(T));
	for(i=0; i<*n; ++i) {
		if(fscanf(file, Tsf, (*t) + i) != 1) {
			free(t);
			return 0;
		}
	}

	return 1;
}

/**
 * Print a given vector using standard error output.
 *
 * \param name Name of the vector (C-style string).
 * \param v Pointer to vector data.
 * \param n Size of the vector.
*/
void print_vector(const char *name, T *v, int n) {
	int i;
	printf("%s:%s", name, n >= 12 ? "\n" : "");
	for(i=0; i<n; ++i) {
		printf("%12.6g ", v[i]);
		if((i+1) % 12 == 0) printf("\n");
	}
	if(n >= 12 && ((i+1) % 12 != 0)) printf("\n");
	printf("\n");
}

/**
 * Print a given matrix using standard error output.
 *
 * \param name Name of the matrix (C-style string).
 * \param m Pointer to matrix data (column ordering).
 * \param rows Number of rows.
 * \param cols Number of columns.
 * \param lda Leading dimension of the matrix.
*/
void print_matrix(const char *name, T *m, int rows, int cols, int lda) {

	int i, j;
	printf("%s:\n", name);
	if(rows <= 0 || cols <= 0 || lda == 0) return;

	for(j=0; j<rows; ++j) {
		for(i=0; i<cols; ++i) printf("%9.5f ", m[i*lda+j]);
		printf("\n");
	}
	printf("\n");
}

